# coding: utf-8

import elementtree.ElementTree as ET
from xml.dom.minidom import parseString

import pygame

import gui

import os
import sys

from widgets import *
from gallox.utils import load_image
from gallox.settings import *

HELP_TEXT = """
gallox map editor

left mouse button - move sprites
middle mouse button / scroll wheel - drag to move map
right  mouse button - resize sprites (resizable sprites only)
scroll wheel up/down - turn sprites (turnable sprites only)
"""

ABOUT_TEXT = """
gallox map editor

by Julian *Jug* Habrock
bytemuehle.de

made with GUI
"""

GENERAL_ATTRIBUTES = [
        ("name", "name", 3, 30, str),
        ("intro message", "message", 0, 100, str),
        ("author1", "author", 0, 100, str),
        ("author2", "author", 0, 100, str),
        ("author3", "author", 0, 100, str)
                    ]

def indent(elem, level=0):
    i = "\n" + level*"  "
    if len(elem):
        if not elem.text or not elem.text.strip():
            elem.text = i + "  "
        for e in elem:
            indent(e, level+1)
            if not e.tail or not e.tail.strip():
                e.tail = i + "  "
        if not e.tail or not e.tail.strip():
            e.tail = i
    else:
        if level and (not elem.tail or not elem.tail.strip()):
            elem.tail = i


class Editor(gui.GUI):
    used_styles=["bgcolor"]
    def __init__(self, size, quit_callback=sys.exit):
        # must be imported after pygame.init()
        import styles
        
        # create interface
        gui.GUI.__init__(self, size, style=gui.Style(**styles.main))
        self.quit_callback = quit_callback
        self.head = gui.Layer(self, ("100%", 100), style=styles.head)
        icons = (
                     ("new", self.handle_new),
                     #! remove lambda
                     ("open", lambda **kw: self.load(self.map_path, **kw)),
                     ("save", self.save),
                     ("save_as", self.dummy_callback),
                     ("view_fullscreen", self.dummy_callback),
                     ("quit", self.quit),
                 )
        x = 20
        for name, callback in icons:
            IconButton(self, pos=(x, 20), name=name, callback=callback,
                        style_class_name="menubutton")
            x += 40
            
        self.panels = PanelList(self.head, size=("100%", 25), pos=(20,"100%"),
                                anchor="bottomleft",
                                style=styles.panel_list, orient=0)
                                
        self.general = AttributeList(self, size=("100%", "100%-200"), pos=(150, 150),
                                     attributes=GENERAL_ATTRIBUTES)
        self.panels.add_panel(self.general, "General")
        
        
        
        self.sprite_editor = SpriteEditor(self, ("100%","100%-120"), (0, 120),
                                anchor="topleft", style=styles.sprite_editor)
        self.panels.add_panel(self.sprite_editor, "Editor", 2)

        
        t = gui.Text(self, "auto", ("50%", 170), ABOUT_TEXT, anchor="midtop")
        self.panels.add_panel(t, "About")
        t = gui.Text(self, "auto", ("50%", 170), HELP_TEXT, anchor="midtop")
        self.panels.add_panel(t, "Help")
        
        self.panels.activate("Editor")
        
        # anonymous dialogues/panels
        self.new_level_dialogue = NewLevelDialogue(self, 
                                              size=("100%", "100%-200"), 
                                              pos=(0, 60), visible=False)
        self.new_level_dialogue.listen("cancelled", 
                                       lambda **kw:self.panels.reactivate())
                    
        self.new_level_dialogue.listen("created_new_level", 
                                        self.handle_new_created)
        
#        self.no_fullscreen_res = None
        # test
        self.level_path = ""
        self.map_path = os.path.abspath(os.path.join(__file__, "../../../testlevel"))
        self.load(self.map_path)

        
    def dummy_callback(self, widget, **kw):
        print "dummy callback for", widget
        
    def load(self, path, **kw):
        print "add AskForSave dialogue"
        tree = ET.parse(os.path.join(path, "level.xml"))
        root = tree.getroot()
        general = root.find("about")
        sprites = root.find("sprites")
        size = root.find("size")
        if general:
            self.general.load(general)
        if sprites:
            self.sprite_editor.load(sprites, size)
        self.level_path = path
        self.reload_images()
        
    def save(self, path=None, **kw):
        path = path or self.level_path
        root = ET.Element("level")
        self.general.save(root)
        self.sprite_editor.save(root)
        indent(root)
        tree = ET.ElementTree(root)

        tree.write(os.path.join(path, "level.xml"), encoding="utf-8")
        print "saved"
        
        
    #! add button for user to reload imgs
    def reload_images(self):
        for name in ("background", "main", "foreground"):
            path = os.path.join(self.level_path, "%s.png" % name)
            if os.path.exists(path):
                img, r = load_image(path, ALPHA_COLOR if name == "main" else None)
            else:
                img = None
            self.style["map_image_%s" % name] = img
        
    def quit(self, **kw):
        print "add AskForSave dialogue"
        self.quit_callback()

        
    def handle_new(self, **kw):
        self.panels.deactivate()
        self.new_level_dialogue.show()
        
    def handle_new_created(self, name, path, size, **kw):
        self.general.reset(name=name)
        self.sprite_editor.reset(size=size)
        self.level_path = path
        self.save()
        self.reload_images()
        self.panels.reactivate()
    
#    def handle_fullscreen(self, **kw):
#        if self.no_fullscreen_res:
#            self.no_fullscreen_res = None
#            pygame.display.set_mode(self.no_fullscreen_res, pygame.RESIZABLE)
#        else:
#            self.no_fullscreen_res = self.size
#            pygame.display.set_mode((0,0))
#        pygame.display.flip()

        
