import os
import math
import weakref
from elementtree.ElementTree import SubElement

import pygame

import gui

from gallox.utils import load_image
from gallox.settings import *



class IconButton(gui.Widget):
    """
    Simply an icon with a mouse-down callback that autoloads its icon and
    can have an overlay to highlight itself when active (by removing it).
    """
    highlight = True
    used_styles = ["image", "overlay", "overlaybg"]
    def __init__(self, parent, size=(32,32), pos=(0,0), name="dummy",
                 callback=None, **kw):
        img = gui.SmoothScaleImageHandler(os.path.join(DATA_PATH, "gui","icons",
                               "%s.png" % name), "alpha")
        gui.Widget.__init__(self, parent, size, pos, style={"image":img},
                            **kw)
        self.name = name
        self.listen("post_mouse_down", callback or self.dummy_callback)
        
    def blit_widget(self, styles):
        if "overlay" in styles and styles["overlay"]:
            obg = styles.get("overlaybg", None)
            if obg:
                self.image.fill(obg)
            drects = gui.Widget.blit_widget(self, styles)
            image = styles["overlay"]
            rect = self.image.get_rect()
            image_rect = image.get_rect(center=rect.center)
            self.image.blit(image, image_rect)
        else:
            drects = gui.Widget.blit_widget(self, styles)
        return drects
    
    def dummy_callback(self, **kw):
        print "no callback defined for <IconButton '%s'>" % self.name
        
#! add to gui.widgets?

class ToggleText(gui.Text):
    """
    A text widget that toggles several texts on mouse_down
    """
    highlight = True
    def __init__(self, parent, size, pos, texts, align="left", **kwargs):
        gui.Text.__init__(self, parent, size, pos, texts[0], align=align, **kwargs)
        self.texts = texts
        self.index = 0
        
    def handle_mouse_down(self, button, **kw):
        ind = self.index
        if button == 1:
            self.index += 1
        elif button == 3:
            self.index += len(self.texts) - 1
        self.index %= len(self.texts)
        if ind != self.index:
            self._send_event("toggled", index=self.index, 
                             text=self.texts[self.index])
        self.text = self.texts[self.index]
        
    def toggle(self):
        # simply simulate a mouse_down
        self.handle_mouse_down(1)

class PanelSwitch(gui.Layer):
    """
    Kind of a tab head button, only useful when combined
    with :class:``PanelList``
    Is either active or not and has a reference (self.panel)
    to the Layer with the actual content.
    ``self.active`` should always be the same as ``self.panel.visible``
    
    Uses the suffix "active" when active.
    """
    style_class_name = "panelswitch"
    highlight = True
    def __init__(self, parent, size, pos, panel, **kwargs):
        self.active = False
        self.panel = panel
        gui.Layer.__init__(self, parent, size, pos, **kwargs)
        self.text = gui.Text(self, "auto", ("50%", "50%"), panel.name, 
                             align="centered", anchor="center",
                             pass_events=True, style = self.style)
        self.text.get_suffix = self.get_suffix
        
    def activate(self):
        self.active = True
        self.panel.show()
        
    def deactivate(self):
        self.active = False
        self.panel.hide()
        
    def get_suffix(self, style_name):
        suff = gui.Layer.get_suffix(self, style_name)
        if not suff and self.active:
            suff = "active"
        return suff
        

class PanelList(gui.Layer):
    """
    Manages a list of Layers like tabs (tab bodies). 
    Tab heads (PanelSwitches) may be horizontally (0)
    or vertically (1) oriented. 
    
    Add new panels with self.add_panel
    
    Use deactivate to hide all panels and reactivate
    to show the last active panel again.
    """
    def __init__(self, parent, size=None, pos=(0,0), orient=0, padding=10,tabsize=(90,25), **kwargs):
        gui.Layer.__init__(self, parent, size, pos, **kwargs)
        self.panels = []
        self.switches = {}
        self.active_panel = None
        self.orient = orient
        self.padding = padding
        self.tabsize = tabsize
        self.last_active_panel = None
        
    def get_styles(self):
        styles = gui.Layer.get_styles(self)
        
        return styles
    
    def add_panel(self, panel, name, index="end"):
        """
        :type panel: gui.Layer
        :param panel: a new panel
        :type name: str
        :param name: 
            name that will be used as tab text and also
            for internal reference
        :type index: int or "end"
        :param index:
            index for the new panel
            
        Add new panel to the PanelList
        """
        panel.name = name
        if index == "end":
            self.panels.append(panel)
        else:
            self.panels.insert(index, panel)
        self.reorder_panels()
        self.activate(name)
        
    #! add remove_panel(name)
        
    def reorder_panels(self):
        self.widgets = []
        self.switches = {}
        pos = [0, 0]
        for panel in self.panels:
            p = PanelSwitch(self, self.tabsize, pos, panel,
                            style=self.style)
            p.listen("mouse_down", self.handle_ps_clicked)
            self.switches[panel.name] = p
            pos[self.orient] += p.size[self.orient] + self.padding
            
            
    def handle_ps_clicked(self, widget, **kw):
        if self.active_panel:
            self.active_panel.deactivate()
        self.active_panel = widget
        self.active_panel.activate()
        
    def activate(self, name):
        """
        activate panel with name ``name``
        """
        self.handle_ps_clicked(self.switches[name])
        
    def deactivate(self):
        """
        deactivate the currently active panel without activating
        another panel and remember it.
        """
        if self.active_panel:
            self.last_active_panel = self.active_panel.panel.name
            self.active_panel.deactivate()
        self.active_panel = None
        
    def reactivate(self):
        """
        activate the last active panel
        """
        self.activate(self.last_active_panel or self.panels[0].name)
        self.last_active_panel = None
        
    
class AttributeList(gui.Layer):
    """
    List of general level attributes such as name, 
    intro message, etc. (see gallox.editor.GENERAL_ATTRIBUTES)
    
    """
    def __init__(self, parent, size, pos, attributes, **kwargs):
        gui.Layer.__init__(self, parent, size, pos, **kwargs)
        x, y = 10, 30
        self.inputs = {"authors":[]}
        for name, keyword, minlenght, maxlength, type in attributes:
            i = gui.TextInput(self, "auto", (x+80, y), 
                            "", template="%s:  %%s" % name, multiline=False, 
                            align="left", anchor="topleft")
            y += i.size[1] * 1.5
            if keyword != "author":
                self.inputs[keyword] = i
            else:
                self.inputs["authors"].append(i)
            
    def reset(self, name=""):
        """
        clear all input fields, then set name to ``name``
        """
        for keyword, input in self.inputs.iteritems():
            if keyword == "name":
                input.set_text(name)
            elif keyword == "authors":
                for i in input:
                    i.set_text("")
            else:
                input.set_text("")
            
    def load(self, tree):
        """
        :type tree: Element
        :param tree: subtree with general attributes
        """
        for keyword, input in self.inputs.iteritems():
            if keyword != "authors":
                text = tree.findtext(keyword)
                if text:
                    input.set_text(text)
            else:
                authors = tree.findall("author")
                for i in input:
                    if authors:
                        author = authors.pop(0)
                        if len(author):
                            name = author.findtext("name")
                            if name:
                                i.set_text(name)
                    else:
                        break
                    
                    
    def save(self, parent):
        """
        :type tree: Element
        :param tree: parent to add the "about" subtree to
        """
        general = SubElement(parent, "about")
        for keyword, input in self.inputs.iteritems():
            if keyword != "authors":
                SubElement(general, keyword).text = input.get_text()
            else:
                for i in input:
                    name = i.get_text()
                    if name:
                        author = SubElement(general, "author")
                        SubElement(author, "name").text = name
                        

MAX_FILENAME_LENGTH = 18
    
#! make more generic and add to gui.widgets

class FileBrowser(gui.Window):
    """
    A file browser window.
    
    select either an image, a directory or a level
    """
    style_class_name = "window"
    image_exts = [".png", ".gif", ".jpg", ".jpeg", ".bmp", ".tiff"]
    def __init__(self, parent, size, pos, filter, callback, msg="", **kwargs):
        """
        :param filter:
            either "level", "dir" or "image"
        """
        self.init = False
        gui.Window.__init__(self, parent, "file browser", size, pos, **kwargs)
        self.path = os.path.abspath(os.getcwd())        
        self.main_path = gui.TextInput(self, "auto", (20,20), text=self.path,
                                        multiline=False)
        self.main_path.listen("pressed_return", 
                                lambda **kw: self.get_gui().set_focus(None))
        self.main_path.listen("focus_lost", self.handle_new_path)
        self.dir_box = gui.ScrolledLayer(self, ("100%-20", "100%-150"), (10, 70))
        self.scrollbar = gui.Scrollbar(self.dir_box,                            
                                       ("100%", 20), (0, "100%"),
                                       style={"borderwidth": (15, 7)},
                                       orient="x", anchor="bottomleft")
        if filter == "dir":
            self.ok_but = gui.Text(self, "auto", (20, "100%-20"), "ok",
                                    anchor="bottomleft", highlight=True) 
            self.ok_but.listen("mouse_down", self.handle_ok)
        self.cancel_but = gui.Text(self, "auto", ("100%-20", "100%-20"), 
                                   "cancel", anchor="bottomright",
                                   highlight=True) 
        self.cancel_but.listen("mouse_down", self.handle_cancel)
        
        self.filter = filter
        self.callback = callback
        self.build_dir_info()
        self.init = True
        
        
    def handle_new_path(self, **kw):
        path = self.main_path.get_text()
        if os.path.isdir(os.path.abspath(path)):
            self.path = path
            self.build_dir_info()
        else:
            self.main_path.set_text(self.path)
            
    def handle_callback(self, widget, **kw):
        if self.filter == "level":
            path = os.path.join(self.path, widget.full_text, "level.xml")
        elif self.filter == "image":
            path = os.path.join(self.path, widget.full_text)
        self.delete()
        self.callback(path)
            
    def build_dir_info(self):
        # remove all children but the scrollbar
        self.dir_box.widgets = [self.scrollbar]
        box_height = self.dir_box.size[1]
        # get a case-insensitive sorted list of subdirectories
        dir_list = [(x.lower(), x) for x in os.listdir(self.path)]
        dir_list.sort()
        
        x,y = 0, 0
        for name in [".."] + [names[1] for names in dir_list]:
            abspath = os.path.abspath(os.path.join(self.path, name))
            special = False
            if not name.startswith(".") or name == "..":
                if os.path.isdir(abspath):
                    if (self.filter == "level" and 
                        os.path.exists(os.path.join(abspath, "level.xml"))):
                        special = True
                elif (self.filter == "image" and 
                      os.path.splitext(abspath)[1] in self.image_exts):
                      special = True
                else:
                    continue
                    
                if y > box_height-50:
                    y = 0
                    x += 150
                    
                if len(name) > MAX_FILENAME_LENGTH:
                    n = MAX_FILENAME_LENGTH - 5
                    start = name
                    tail = ""
                    if self.filter == "image" and special:
                        start, tail = os.path.splitext(name)
                        n -= len(tail)
                    short_name = start[:n]+"[...]"+tail
                else:
                    short_name = name
                t = gui.Text(self.dir_box, "auto", (x,y), short_name,align="left")
                t.full_text = name
                if special:
                    t.template = "<%s>"
                    t.listen("mouse_down", self.handle_callback)
                else:
                    t.listen("mouse_down", self.handle_dir_clicked)
                y += t.size[1] * 1.5
                
        self.dir_box.scroll_size = (x+150, 0)
        self.dir_box.init_pos()
        self.dir_box.dirty = True
                
    def handle_cancel(self, **kw):
        self._send_event("cancelled")
        self.delete()
    
    def handle_ok(self, **kw):
        if not self.filter == "dir":
            return
        self.delete()
        self.callback(self.path)
            
    def handle_dir_clicked(self, widget, **kw):
        self.path = os.path.abspath(os.path.join(self.path, widget.full_text))
        self.main_path.set_text(self.path)
        self.build_dir_info()
    
    def get_styles(self):
        styles = gui.Window.get_styles(self)
        styles["path"] = self.path
        return styles
    
    def handle_resized(self, **kw):
        if self.init:
            self.build_dir_info()
            
            
    
DEFAULT_ATTRS = {"fix_no":False, "turnable":False, "resizable":False}
SPRITES = {
        "plattform":{"resizable":True},
        "gun":{"turnable":True},
        "ship":{"fix_no":True, "turnable":True},
           }
           
           
def mybool(s):
    """
    convert a string to a boolean and also consider "False" as False
    
    this is neccessary to save booleans with str(boolean) explicitly
    as "True" and "False". Otherwise False would have to be saved as
    "" what is too implicit.    
    """
    return s not in ("", "False")

# parameter types
BOOL = 1
INT = 2
FLOAT = 3
# list of functions to convert a value from a string back to its original type
CONVERT_FUNCTIONS = [None, mybool, int, float]
           
"""
General information about the parameters a sprite of a certain type
can have, structured as {type : [list of params]} dict where one 
parameter consists of 

  - a name, that is used for any internal usage
  - a repr_name, that is used whenever the name is displayed for the user
  - a type, that must be one of BOOL, INT, FLOAT
  - a default value
  - an args dict, that will be passed to the widget used for this type.
    
Two useful values for the arg dict of an int param are, for example, "min_val"
and "max_val" (used by gui.NumberInput) to control the range of the param value.
    
"""
SPRITE_PARAMS = {
        "plattform":[
               { "name": "refuel",
                 "repr_name": "refuel",
                 "type": BOOL,
                 "default": True,
                 "args": {}
                },
                    ],
        "gun":[
               {"name": "actionScope",
                "repr_name": "action scope",
                "type": INT,
                "default":60,
                "args": {"max_val":361,
                         "min_val":0}
                },
                {"name": "autoAim",
                 "repr_name": "auto aim",
                 "type": BOOL,
                 "default": False,
                 "args": {}
                },
                {"name":"rotSpeed",
                "repr_name":"rotary speed [deg/s]",
                "type":FLOAT,
                "default": 4,
                "args":{"max_val":100,
                        "min_val":0}
                },
                {"name":"shotFreq",
                "repr_name":"shot frequency [px/s]",
                "type":FLOAT,
                "default": 5,
                "args":{"max_val":100,
                        "min_val":1}
                },
               ],
        "ship":[],
           }

#! do not define such values here
HIGHLIGHT_COLOR = (255,0, 200)
                                             
class Sprite(gui.Widget):
    """
    Represents game sprites of any type.
    
    sprite.attributes 
    ''''''''''''''''''
    depend on the sprite type and influence the way the
    sprite can be edited, ie. if it can be resized or turned.
    
    
    sprite.parameters
    ''''''''''''''''''
    depend on the individual object and have only influence in
    the actual game and can be edited via the param tab of the
    sprite editor.

    
    names
    '''''''''''
    Every sprite object has a representative, unique name, 
    sprite.repr build from the type and a serial number that
    makes it easy to reference sprites. Widgets that belong
    to a certain sprite need to save only the repr of a sprite.
    To get the sprite to a given repr use the :meth:``Sprite.get``
    class method.
    """
    counters = {}
    _id2obj_dict = weakref.WeakValueDictionary()
    def __init__(self, parent, name, size=None, pos=(0,0),degrees=0, **kwargs):
        if name not in SPRITES:
            raise ValueError("unknown sprite '%s'" % name)
        
        self.name = name
        #self.can_turn = SPRITES[name].get("can_turn", False)
        self.degrees = degrees
        # what a sprite of this type can do 
        self.attributes = dict(DEFAULT_ATTRS)
        self.attributes.update(SPRITES[name])
        
        # individual parameters for this single sprite object
        # first, get names and default values
        self.parameters = dict([(d["name"], d["default"]) 
                                for d in SPRITE_PARAMS[self.name]])
        
        if self._is("fix_no"):
            if self.name in self.counters:
                raise ValueError, "number of %s sprites limited to 1" % self.name
            else:
                self.repr = u"%s" % self.name
        else:
            if self.name not in self.counters:
                self.counters[self.name] = count = 1
            else:
                self.counters[self.name] += 1
                count = self.counters[self.name]
            self.repr = u"%s %i" % (name, count)
        
        self._id2obj_dict[self.repr] = self        
        
        self.imgh = gui.SmoothScaleImageHandler(os.path.abspath(os.path.join(__file__,
                                "../../../data/sprites/%s.png" % name)), "alpha")
        if not "style" in kwargs:
            kwargs["style"] = {}
        kwargs["style"]["image"] = self.imgh
        if not self._is("resizable"):
            size = None
        size = size or self.imgh.get_original().get_size()
        gui.Widget.__init__(self, parent, size, pos, **kwargs)
        
        
    def __str__(self):
        return "%s" % str(self.repr)
    
    def __unicode__(self):
        return u"%s" % self.repr
    
    def __repr__(self):
        return "%s" % self.repr
    
    @classmethod
    def get(cls, name):
        """
        :type name: str
        :param name:
            repr of a sprite object
        :rtype: :class:`Sprite`
        :returns:
            the sprite object represented by repr
            
        Get the sprite object from a given repr
        """
        try:
            return cls._id2obj_dict[str(name)]
        except KeyError:
            raise KeyError, "%s; valid keys are: %s" % (name, 
                                        ", ".join(cls._id2obj_dict.iterkeys()))
        
    def handle_mouse_dragged(self, buttons, rel, abs, **kwargs):
        """
        move the sprite with left mouse button pressed or resize it with
        right button pressed
        """
        if buttons[0]:
            self.move(rel)
        elif buttons[2] and self._is("resizable"):
            if self._is("turnable"):
                x,y = abs
                abs = x+y, x+y
            self.resize(abs)
            
    def handle_mouse_down(self, button, **kwargs):
        """
        rotate the sprite by scrolling (if allowed)
        """
        if self._is("turnable"):
            if button == 4:
                self.degrees += 10
            elif button == 5:
                self.degrees -= 10
            
    def blit_widget(self, styles):
        if "image" in styles and styles["image"]:
            image = pygame.transform.rotozoom(styles["image"], self.degrees, 1)
            rect = self.image.get_rect()
            image_rect = image.get_rect(center=rect.center)
            self.image.blit(image, image_rect)
            if self.focused:
                pygame.draw.rect(self.image, HIGHLIGHT_COLOR, 
                                 rect.inflate(-1,-1), 2)
            return [self.rect]
        
    def _is(self, attr):
        return self.attributes[attr]
        
    def get_styles(self):
        styles = gui.Widget.get_styles(self)
        styles["degrees"] = self.degrees
        styles["focused"] = self.focused
        return styles
    
class Map(gui.ScrolledLayer):
    """
    The actual map.
    Scrollable Layer that is parent of all Sprites.
    """
    used_styles = ["image_background", "image_main", "image_foreground"]
    style_class_name = "map"
    def __init__(self, parent, size=None, pos=(0,0), **kw):
        gui.ScrolledLayer.__init__(self, parent, size, pos=pos, **kw)
        self.init = False
        self.scrollbars_x = [gui.Scrollbar(self, ("100%", 15), (0, 0), orient="x", 
                      anchor="topleft", style={"borderwidth": (15, 4)}),
                      gui.Scrollbar(self, ("100%", 15), (0, "100%"), orient="x", 
                      anchor="bottomleft", style={"borderwidth": (15, 4)})]
        self.scrollbars_y = [gui.Scrollbar(self, (15, "100%"), (0, 0), orient="y", 
                      anchor="topleft", style={"borderwidth": (4, 15)}),
                      gui.Scrollbar(self, (15, "100%"), ("100%", 0), orient="y", 
                      anchor="topright", style={"borderwidth": (4, 15)})]
        self.auto_dragging = False
        self.init = True
       
    def _get_sprites(self):
        return [widget for widget in self.widgets if not isinstance(widget, gui.Scrollbar)]

    sprites = property(_get_sprites)
        
    def blit_widget(self, styles):
        gui.ScrolledLayer.blit_widget(self, styles)
        any_image = False
        for scrollbar in self.scrollbars_x + self.scrollbars_y:
            scrollbar.update_slider_size()
        for name in self.used_styles:
            if name in styles and styles[name]:
                image = styles[name]
                #self.image.blit(image, (self.offset_x, self.offset_y))
                rect = self.get_rect(self)
                rect.topleft = -self.offset_x, -self.offset_y
                bd = self.style("borderwidth", self, default=0)
                if type(bd) == int:
                    x = y = bd
                else:
                    x,y = bd
                self.image.blit(image, (x,y), rect)
                any_image = True
        if any_image:
            return [self.rect]
        
    def blit_widget_top(self, styles):
        if self.dragging[1] and self.dragging_anchor and self.last_drag_pos:
            # draw scrolling controller
            x1, y1 = self.dragging_anchor
            x2, y2 = self.last_drag_pos
            my_x, my_y = self.rect.topleft
            x2 -= my_x
            y2 -= my_y
            
            
            drag_rect = pygame.Rect(0, 0, 20, 20)
            drag_rect.center = self.dragging_anchor

            start_pos = [[drag_rect.midleft, drag_rect.midright], 
                         [drag_rect.midtop, drag_rect.midbottom]]
            points = [drag_rect.topleft, drag_rect.bottomleft,
                      drag_rect.bottomright, drag_rect.topright,]
#            start_pos = [[points[0], points[3]],[points[1], points[2]]]
#            for point1, point2 in start_pos:
#                pygame.draw.line(self.image, (0,0,0), point1, point2, 3)
            pygame.draw.lines(self.image, (0,0,0), 1, points, 3)
            colors = [(200,0,0), (0,200,0), (0,0,200), (200,200,0)]
            for point, color in zip(points, colors):
                pygame.draw.line(self.image, color, point, (x2,y2), 3)
            
            
    def handle_new_scroll_size(self, **kw):
        sx, sy = self.scroll_size
        bd = self.style("borderwidth", self, default=0)
        if type(bd) == int:
            bx = by = bd
        else:
            bx, by = bd
        self.max_size = sx + bx * 2, sy + by * 2
        
    def get_rect(self, widget):
        rect = gui.ScrolledLayer.get_rect(self, widget)
        if not self.init:
            return rect
        if widget in self.scrollbars_x:
            rect.height = self.rect.height
            rect.center = self.rect.center
        if widget in self.scrollbars_y:
            rect.width = self.rect.width
            rect.center = self.rect.center
        return rect
    
    def get_visible_center(self):
        return (- self.offset_x + self.size[0]/2., 
                - self.offset_y + self.size[1]/2.)

    def clear(self):
        # remove all widgets but scrollbars
        self.widgets = self.fixed_widgets[:]
        
    def handle_mouse_dragged(self, rel, abs, buttons, **kw):
        if buttons[1]:
            self.scroll_x(-rel[0]/2.)
            self.scroll_y(-rel[1]/2.)
#            self.mouse_scrolling = True
            
    def handle_mouse_stopped_dragging(self, **kw):
        # clear scrolling line
        self.dirty = True
#        self.mouse_scrolling = False

#    def handle_widget_added(self, new_widget, **kw):
#        new_widget.listen("mouse_dragged", 
#                          self.handle_mouse_dragged)
#        new_widget.listen("mouse_stopped_dragging", 
#                          self.handle_mouse_stopped_dragging)
        
    def handle_mouse_left(self, last_pos, **kw):
        if any(pygame.mouse.get_pressed()) and not self.dragging[1]:
            self.dragging_anchor = last_pos[0]-self.rect.top, last_pos[1]-self.rect.left
#            print "dragging", self.dragging_anchor
            pygame.event.post(pygame.event.Event(pygame.MOUSEBUTTONDOWN, button=1, pos=last_pos))
            self.auto_dragging = True

    def handle_mouse_entered(self, **kw):
        if self.auto_dragging:
            self.auto_dragging = False
            self.dragging[1] = False


        
class List(gui.Layer):
    """
    List of items (names/strings) with additional :class:``IconButton``\ s
    
    #! todo
    self.meta_objects -> don't remove
    like "--- main layer ---"
    """
    def __init__(self, parent, size=None, pos=(0,0), items=[], icons=[], 
                 events=[], item_kwargs={}, **kwargs):
        """
        :type items: list
        :param items: list item names
        :type icons: 2d list
        :param icons: 
            list of (icon name, callback) pairs
            ``callback`` may be a callback function for mouse_down events
            or an (even name, callback function) pair
        :type events: 2d list
        :param events: 
            list of (event name, callback) pairs to be applied to
            the items
        :type item_kwargs: dict
        :param item_kwargs:
            key word parameters that will be passed to item text widgets
        """
        gui.Layer.__init__(self, parent, size, pos, **kwargs)
        self.icons = icons
        self.item_events = events
        self.item_kwargs = item_kwargs
        self.update_items(items)
        self.lines = {}
        self.item_width
        self.highlight_item = None
        
    def blit_widget(self, styles):
        gui.Layer.blit_widget(self, styles)
        if self.highlight_item:
            widget = self.lines[self.highlight_item]["text"]
            wx, wy = widget.local_pos
            rect = pygame.Rect(0, wy - 5, self.item_width, widget.size[1] + 10)
            self.image.fill((72, 224, 219, 170), rect)
            
            pygame.draw.line(self.image, (0,0,0), rect.topleft, rect.topright, 2)
            pygame.draw.line(self.image, (0,0,0), rect.bottomleft, rect.bottomright, 2)
        
    def update_items(self, items):
        def item_cb(item, callback):
            def cb(**kw):
                callback(item=item, **kw)
            return cb
        def dummy(**kw):
            pass
        
        #remove old widgets
        self.widgets = []
        self.lines = {}
        x,y = 0, 0
        max_text_width = 0
        for name in items:
            t = gui.Text(self, "auto", (0,y), str(name), **self.item_kwargs)
            h = t.size[1]
            max_text_width = max(t.size[0], max_text_width)
            y += h * 1.5
            self.lines[name] = {"text": t}
            
        y = 0
        for name in items:
            x = max_text_width + 10
            for event, callback in self.item_events:
                
                self.lines[name]["text"].listen(event, item_cb(name, callback))
                
            #x += t.size[0] + 20
            for icon, callback in self.icons:
                if callable(callback):
                    i = IconButton(self, (h,h), (x,y), icon, 
                         item_cb(name, callback), style_class_name="editoribut")
                else:
                    # this way, its possible to define multiple callbacks for
                    # one icon
                    if icon in self.lines[name]:
                        i = self.lines[name][icon]
                        x -= i.size[0] + 10
                    else:
                        i = IconButton(self, (h,h), (x,y), icon, dummy, 
                                        style_class_name="editoribut")
                    event_name, callback = callback
                    i.listen(event_name, item_cb(name, callback))
                x += i.size[0] + 10
                self.lines[name][icon] = i
            y += h * 1.5
        self.item_width = x
        self.dirty = True
        
    def get_styles(self):
        styles = gui.Layer.get_styles(self)
        styles["highlight_item"] = self.highlight_item
        return styles


class SpriteParamEditor(gui.Layer):
    """
    Editor for sprite parameters.
    Set self.sprite to any sprite to make its parameters editable.
    Gets all information from SPRITE_PARAMETERS.
    """
    def __init__(self, parent, size, pos, **kwargs):
        gui.Layer.__init__(self, parent, size, pos, **kwargs)
        self._sprite = None
    
    def get_sprite(self):
        return self._sprite
        
    def set_sprite(self, sprite):
        self._sprite = sprite
        self.rebuild()
        
    sprite = property(get_sprite, set_sprite)
    
    def rebuild(self):
        """
        build interface
        """
        # remove old stuff
        self.widgets = []
        self.dirty = True
        # if not sprite is selected, there is nothing to show
        if self.sprite is None:
            return
        y = 0
        for param in SPRITE_PARAMS[self.sprite.name]:
            if param["type"] == BOOL:
                texts = ["no", "yes"]
                # value = no/0/False
                widget = ToggleText(self, "auto", (0, y), texts, 
                                    template="%s: %%s" % param["repr_name"])
                # toggle if value is yes/1/True
                if self.sprite.parameters[param["name"]]:
                    widget.toggle()
                widget.listen("toggled", self.handle_toggled)
                
            elif param["type"] in [INT, FLOAT]:
                if param["type"] == INT:
                    cf = int
                else:
                    cf = float
                widget = gui.NumberInput(self, "auto", (0,y), 
                               text=str(self.sprite.parameters[param["name"]]),
                               template="%s: %%s" % param["repr_name"], 
                               align="left", convert_func=cf, **param["args"])
                widget.listen("number_changed", self.handle_new_num_val)
                self.sprite.parameters[param["name"]] = widget.get_number()
            else:
                continue
            widget.param_name = param["name"]
            y += widget.size[1]
        
        
    #! maybe update sprite and add some visual hint for new param?
    
    def handle_new_num_val(self, widget, new, **kw):
        self.sprite.parameters[widget.param_name] = new
        
    def handle_toggled(self, widget, index, **kw):
        self.sprite.parameters[widget.param_name] = bool(index)
        
        
        
        
        
    
class SpriteEditor(gui.Layer):
    """
    Main editor component that contains the :class:``Map``\ , 
    the :class:``SpriteParamEditor`` and some :class:``Lists``
    to edit and manage all sprites.
    """
    def __init__(self, parent, size, pos, **kwargs):
        gui.Layer.__init__(self, parent, size, pos, **kwargs)
        style = self.style

        self.map = Map(self, ("100%-270", "100%-20"), (10,10), style=self.style)
        Sprite(self.map, "plattform"),
        Sprite(self.map, "gun", (100,40))
                        
        self.panels = PanelList(self, size=("30%", 40), pos=("100%-250", 0),
                                tabsize=(40,30), style=self.style
                                )
                                
        sidebar_size, sidebar_pos = (250,"100%-50"), ("100%-250", 50)
        self.sprite_list = List(self, sidebar_size, sidebar_pos,
                                items=[k for k in SPRITES if k != "ship"],
                                icons=[("add", self.handle_sprite_add)],
                                events=[("post_mouse_down", self.handle_sprite_select)],
                                item_kwargs={"highlight":True},
                                style=self.style)

        self.panels.add_panel(self.sprite_list, "sprites")

        
        self.obj_list = List(self, sidebar_size, sidebar_pos,
                                items=self.map.sprites,
                                icons=[("move", ("mouse_dragged", self.handle_obj_move)),
                                       ("move", ("mouse_stopped_dragging", self.handle_obj_moved)),
                                       ("preferences", self.handle_obj_prefs),
                                       ("remove", self.handle_obj_remove)],
                                events=[("post_mouse_down", self.handle_obj_select)],
#                                item_kwargs={"highlight":True}
                                style=self.style
                                )

        self.panels.add_panel(self.obj_list, "objects")

        
        self.param_editor = SpriteParamEditor(self, sidebar_size, sidebar_pos, style=self.style)
        
        self.panels.add_panel(self.param_editor, "params")
        
        
        #! specify list (update_list(listX))
        self.map.listen("widget_added", self.update_lists)
        self.map.listen("widget_removed", self.update_lists)
        self.map.listen("widget_index_changed", self.update_lists)
        
    def handle_sprite_select(self, item, **kw):
        # update self.obj_list.
        # set filter (item.name)
        # self.panels.activate(self.obj_list)
        print "show %ss" % item
        
    def handle_sprite_add(self, item, **kw):
        # add new sprite to map
        self.get_gui().set_focus(Sprite(self.map, item, 
                                        pos=self.map.get_visible_center()))
        
        
    def handle_obj_select(self, item, **kw):
        #! highlight in list too
        self.get_gui().set_focus(Sprite.get(item))
    
    def handle_obj_move(self, item, widget, **kw):
        self.obj_list.highlight_item = item
        mouse_pos = pygame.mouse.get_pos()
        if mouse_pos[1] < widget.rect.top - 5:
            self.map.widget_down(Sprite.get(item))
        elif mouse_pos[1] > widget.rect.bottom + 5:
            self.map.widget_up(Sprite.get(item))
        else:
            return
        i = (widget.last_mouse_pos, widget.dragging, widget.dragging_anchor,
             widget.last_drag_pos)
#        self.update_lists()
        widget = self.obj_list.lines[item]["move"]
        widget.last_mouse_pos, widget.dragging, widget.dragging_anchor,\
             widget.last_drag_pos = i
             
#        for y in (widget.rect.top, widget.rect.bottom):
#            pygame.draw.line(self.obj_list.image, (0,0,0), (0,y), 
#                            (self.obj_list.item_width+5, y), 2)
        
    def handle_obj_moved(self, item, **kw):
        """
        dragging stopped, clear highlight
        """
        self.obj_list.highlight_item = None
        
        
    def handle_obj_prefs(self, item, **kw):
        print "obj prefs"
        self.param_editor.sprite = Sprite.get(item)
        self.panels.activate(self.param_editor.name)
        
    def handle_obj_remove(self, item, **kw):
        # remove item from map
        self.map.remove(Sprite.get(item))
        
        
    def update_lists(self, **kw):
        self.obj_list.update_items([str(name) for name in self.map.sprites])
        # ship can not be deleted
        if "ship" in self.obj_list.lines:
            self.obj_list.lines["ship"]["remove"].delete()
        
        
        
        
        
    def load(self, sprite_tree, size_tree):
        """
        :type tree: Element
        :param tree: subtree with general attributes
        """
        #self.sprites = []
        self.map.clear()
        Sprite.counters = {}
        has_ship = False
        self.map_size = (int(size_tree.findtext("width")),
                         int(size_tree.findtext("height")))
        self.map.scroll_size = self.map_size
        for se in sprite_tree.findall("sprite"):
            name = se.get("type")
            p = se.find("position")
            pos = int(p.get("x")), int(p.get("y"))
            s = se.find("size")
            size = int(s.get("w")), int(s.get("h"))
            degrees = int(se.get("degrees", 0))
            if not name == "ship" or not has_ship:
                s = Sprite(self.map, name, size=size, pos=pos, degrees=degrees)
                if name == "ship":
                    has_ship = True
                for param in SPRITE_PARAMS[name]:
                    #! improve
                    cf = CONVERT_FUNCTIONS[param["type"]]
                    val = se.findtext(param["name"])
                    if val:
                        s.parameters[param["name"]] = cf(val)
        if not has_ship:
            Sprite(self.map, "ship")
        self.map.dirty = True
        # may still refference to an old sprite
        self.param_editor.sprite = None
        self.update_lists()
                    
    def save(self, parent):
        size = SubElement(parent, "size")
        SubElement(size, "width").text = str(self.map_size[0])
        SubElement(size, "height").text = str(self.map_size[1])
        general = SubElement(parent, "sprites")
        for sprite in self.map.sprites:
            se = SubElement(general, "sprite", type=sprite.name)
            if sprite._is("turnable"):
                se.attrib["degrees"] = str(sprite.degrees)
            SubElement(se, "position", x=str(sprite.local_pos[0]), y=str(sprite.local_pos[1]))
            SubElement(se, "size", w=str(sprite.size[0]), h=str(sprite.size[1]))
            for p_name, p_val in sprite.parameters.iteritems():
                SubElement(se, p_name).text = str(p_val)
            
    def reset(self, size):
        self.map.clear()
        Sprite.counters = {}
        self.map_size = size
        self.map.scroll_size = self.map_size
        Sprite(self.map, "ship", pos=(30,30))
        self.map.dirty = True
        self.update_lists()
        
class NewLevelDialogue(gui.Layer):
    """
    Dialogue to get all data from the user that is required
    to create a new level
    """
    def __init__(self, parent, size, pos, **kwargs):
        gui.Layer.__init__(self, parent, size, pos, **kwargs)
        y = 130
        self.title = gui.Text(self, "auto", ("50%", y), "Create New Level",
                             anchor="center")
        y += 50
        self.name_inp = gui.TextInput(self, "auto", ("50%", y), "NewLevel", 
                                    template="name: %s",
                                    anchor="center")
        y += 40
        self.width_inp = gui.NumberInput(self, "auto", ("50%-50", y), "1000", 
                                    template="width: %s", min_val=300, 
                                    max_val=2000, convert_func=int,
                                    anchor="center")
        self.height_inp = gui.NumberInput(self, "auto", ("50%+50", y), "800", 
                                    template="height: %s", min_val=300, 
                                    max_val=2000, convert_func=int,
                                    anchor="center")
                                    
        y += 30
        self.working_dir_inp = gui.TextInput(self, "auto", ("50%", y), "", 
                                    template="working directory: %s",
                                    anchor="center")
        self.working_dir_inp.listen("mouse_down", self.handle_browse)
        
        y += 50
        self.ok_but = gui.Text(self, "auto", ("50%-50", y), "create")
        self.ok_but.listen("mouse_down", self.handle_create)
        self.cancel_but = gui.Text(self, "auto", ("50%+50", y), "cancel")
        self.cancel_but.listen("mouse_down", self.handle_cancel)
        
    def handle_browse(self, **kw):
        FileBrowser(self.get_gui(), (300,400), (200,50),
                    "dir", self.handle_working_dir)
                    
    def handle_working_dir(self, path, **kw):
        self.working_dir_inp.text = path
                    
    
    def handle_create(self, **kw):
        if not self.name_inp.text:
            print "no level name"
            return
        if not os.path.isdir(self.working_dir_inp.text):
            print "no working dir"
            return
        name = self.name_inp.text
        path = os.path.join(self.working_dir_inp.text, name.lower())
        #! check if exists
        os.mkdir(path)
        width = self.width_inp.get_number()
        height = self.height_inp.get_number()
        bg = pygame.Surface((width, height))
        main = bg.copy()
        bg.fill((200,200,200))
        text = pygame.font.SysFont(None, 60).render("background ", 1, 
                                                (230,230,230), (200,200,200))
        i = 0
        for y in xrange(0, height, text.get_size()[1]):
            for x in xrange(i*-text.get_size()[0]/2, width, text.get_size()[0]):
                bg.blit(text, (x,y))
            i = 1-i
                
        main.fill((0,0,0))
        w, h = main.get_size()
        main.fill(ALPHA_COLOR, (20,20, w-40, h-40))
        text = pygame.font.SysFont(None, 150).render("main level layer", 0,
                                                    (0,0,0), ALPHA_COLOR)
        main.blit(text, text.get_rect(center=(w/2, h/2)))
        pygame.image.save(bg, os.path.join(path, "background.png"))
        pygame.image.save(main, os.path.join(path, "main.png"))
        self._send_event("created_new_level", size=(width, height), name=name,
                          path=path)
        self.hide()
            
        
    
    def handle_cancel(self, **kw):
        self.hide()
        self._send_event("cancelled")
                                    